#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <err.h>

#define SHELL "/bin/sh"

#ifdef PREFIX
#   define JAILUPRC PREFIX"/etc/jailup.rc"
#   define STARTUPDIR PREFIX"/etc/jailup"
#endif

#define MAXRCLEN 0x7FFFFFFE
#define todo     (argv[0])
#define genclass (argv[1])
#define inclass  (argv[2])
#define jrabl 001
#define jrsto 002
#define jrsta 004
#define jrest 010

#define NAMELENMAX 128
char *prefix, *startupdir, *jailuprc;
char exclusives[][NAMELENMAX] = {"syslogd", "named", NULL};
int flag_v = 0;
char *rc;
struct sparced
{   char *gencl;
    char *incl;
    char *tail;
    char mode;
} sparced;

int exclude(char *name)
{   int i;
    for (i = 0; exclusives[i] && exclusives[i][0]; i++)
        if  (0 == strncmp(name, exclusives[i], NAMELENMAX)) return(i + 1);
    return(0);
}

void exup(char *tofun, char *class, char *tail)
{   char *parm;

    if  (flag_v > 1) fprintf(stderr, "Exup: ");
    if  (0 > asprintf(&parm, "%s/%s/%s", startupdir, tofun, class))              err(EX_TEMPFAIL, NULL);
    if  (0 > access(parm, X_OK))                                             err(EX_UNAVAILABLE, NULL);
    free(parm);
    if  (0 > asprintf(&parm, "%s/%s/%s %s&", startupdir, tofun, class, tail))     err(EX_TEMPFAIL, NULL);
    if  (flag_v) fprintf(stderr, "%s\n", parm);
    switch (fork())
    {case -1:                                                                      err(EX_OSERR, NULL);
     case 0:
        if  (0 > daemon(0,1))                                                      err(EX_OSERR, NULL);
        execl(SHELL, "sh", "-c", parm, (char *)NULL);
        syslog(LOG_ERR | LOG_DAEMON, "No %s/%s up: %m", tofun, class);
        _exit(EX_OSERR);
    }
    free(parm);
}

void usage(int ex, char *jailuprc, char *startupdir)
{   int c;

    fprintf( stderr
           , "jailupctl @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "jailupctl [-v[v[v]]] {start|restart|stop} [genclass [inclass]]\n"
             "   -v verbose\n"
             "   reads %s string by string\n"
             "   and starts daemons %s/{start|restart|stop}/{inclass}\n"
             "   PREFIX, JAILUPRC, STARTUPDIR environment variables used\n"
             "   JAILUPRC file must be not longer then %d bytes\n"
             "Exclusives:\n"
           , jailuprc
           , startupdir
           , MAXRCLEN
           );
    for (c = 0; exclusives[c] && exclusives[c][0]; c++)
        fprintf(stderr, "    %d %s\n", c + 1, exclusives[c]);
    exit(ex);
}

char *next(char *in, char *end, struct sparced *sped)
{   unsigned ctl = 0;
    int c, i, j;
    enum {inii, lkgn, dogn, lkin, doin, lktl, dotl, skip, purg, done} state;
    enum symclass {snl, ssp, slt, scm, sxx};
    unsigned char chtab[257] = { snl
        /* 0    1    2    3     4    5    6    7     8    9    A    B     C    D    E    F */
  /*0*/ , snl, slt, slt, slt,  slt, slt, slt, slt,  slt, ssp, snl, slt,  slt, snl, slt, slt
  /*1*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*2*/ , ssp, slt, slt, scm,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*3*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*4*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*5*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*6*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*7*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*8*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*9*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*A*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*B*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*C*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*D*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*E*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
  /*F*/ , slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt,  slt, slt, slt, slt
        };
#   define retu 0000400
#   define popl 0001000
#   define nlfy 0002000
#   define togn 0004000
#   define toin 0010000
#   define totl 0020000
#   define foff 0040000
#   define fonn 0100000
#   define inil 0200000
    unsigned control[done][sxx] =
        /*               snl                  ssp                         slt                  scm*/
/*inii*/{        popl |  inii,  foff | popl | lkgn,  fonn | togn | popl | dogn,  inil | popl | purg
/*lkgn*/, inil | popl |  inii,         popl | lkgn,         togn | popl | dogn,  inil | popl | purg
/*dogn*/, inil | popl |  inii,  nlfy | popl | lkin,                popl | dogn,  inil | popl | purg
/*lkin*/, inil | popl |  inii,         popl | lkin,         toin | popl | doin,  inil | popl | purg
/*doin*/, nlfy | popl | retu ,  nlfy | popl | lktl,                popl | doin,  nlfy | popl | skip
/*lktl*/,        popl | retu ,         popl | lktl,         totl | popl | dotl,         popl | skip
/*dotl*/, nlfy | popl | retu ,         popl | dotl,                popl | dotl,         popl | dotl
/*skip*/,        popl | retu ,         popl | skip,                popl | skip,         popl | skip
/*purg*/,        popl |  inii,         popl | purg,                popl | purg,         popl | purg
        };
    {sped->tail = NULL; sped->incl = NULL; sped->gencl = NULL;};
    for (state = inii; (ctl & retu) == 0; state = ctl & 0xFF)
    {   c = (in < end) ? *in & 0xFF : -1;
        ctl = control[state][chtab[c + 1]];
        if  (flag_v > 2)
            fprintf( stderr, "s%d, c%02X(%d), ctl %08o in=%08X\n"
                   , state, c, chtab[c + 1], ctl, (unsigned)in
                   );
        if  (ctl & inil) {sped->tail = NULL; sped->incl = NULL; sped->gencl = NULL;};
        if  (ctl & fonn) sped->mode = jrabl;
        if  (ctl & foff) sped->mode = 0;
        if  (ctl & totl) sped->tail = in;
        if  (ctl & toin) sped->incl = in;
        if  (ctl & togn) sped->gencl = in;
        if  (ctl & nlfy) *in = '\0';
        if  (ctl & popl) in++;
        if  (c < 0) break;
    };
    return(in);
}

int main(int argc, char **argv, char **envp)
{   int rcfile, c;
    char *in, *p;
    unsigned doma;
    struct stat sb;

    prefix = getenv("PREFIX");
    jailuprc = getenv("JAILUPRC");
    startupdir = getenv("STARTUPDIR");
#ifdef PREFIX
    if  (NULL == prefix) prefix = PREFIX;
#endif
#ifdef JAILUPRC
    if  (NULL == jailuprc) jailuprc = JAILUPRC;
#else
    if  (  NULL == jailuprc
        && NULL != prefix
        && 0 > asprintf(&jailuprc, "%s/etc/jailup.rc", prefix)
        )                                                                       err(EX_TEMPFAIL, NULL);
#endif
#ifdef STARTUPDIR
    if  (NULL == startupdir) startupdir = STARTUPDIR;
#else
    if  (  NULL == startupdir
        && NULL != prefix
        && 0 > asprintf(&startupdir, "%s/etc/jailup", prefix)
        )                                                                       err(EX_TEMPFAIL, NULL);
#endif
    while((c = getopt(argc, argv, "vh?")) > 0)
    {   switch(c)
        {case 'v': flag_v++;                 break;
         case 'h': usage(EX_OK, jailuprc, startupdir);
         default: usage(EX_USAGE, jailuprc, startupdir);
    }   };
    argc -= optind;
    argv += optind;

    if  (NULL == jailuprc || NULL == startupdir || argc < 1 || argc > 3)
                                                                 usage(EX_USAGE, jailuprc, startupdir);
    if  (flag_v) fprintf( stderr, "PREFIX=%s\nJAILUPRC=%s\nSTARTUPDIR=%s\n"
                        , prefix, jailuprc, startupdir
                        );
    if  (flag_v > 1)
    {   fprintf(stderr, "Exclusives:\n");
        for (c = 0; exclusives[c] && exclusives[c][0]; c++)
            fprintf(stderr, "    %d %s\n", c + 1, exclusives[c]);
    }
    doma = 0;
    if  (0 == strcmp(todo, "start")) doma = jrsta;
    if  (0 == strcmp(todo, "stop")) doma = jrsto;
    if  (0 == strcmp(todo, "restart")) doma = jrsto | jrsta;
    if  (flag_v) fprintf( stderr, "Todo %s %03o\n", todo, doma);
    if  (doma == 0)                                              usage(EX_USAGE, jailuprc, startupdir);
    /* Environment and arguments parse ready */

    if  (0 > (rcfile = open(jailuprc, O_RDONLY | O_EXLOCK)))                     err(EX_NOINPUT, NULL);
    if  (0 > fstat(rcfile, &sb))                                                 err(EX_NOINPUT, NULL);
    if  (sb.st_size > MAXRCLEN)                                              err(EX_UNAVAILABLE, NULL);
#ifdef READ_INSTEAD
    if  (NULL == (in = malloc(sb.st_size + 1)))                              err(EX_UNAVAILABLE, NULL);
    /*            VXXXXXXXXV        */
    if  (sb.st_size != read(rcfile, in, sb.st_size))                               err(EX_IOERR, NULL);
#else
    if  (MAP_FAILED == (in = mmap( NULL, sb.st_size + 1
                                 , PROT_READ | PROT_WRITE, MAP_PRIVATE
                                 , rcfile, (off_t)0
        )              )         )                                                 err(EX_IOERR, NULL);
#endif
    close(rcfile);
    for (p = in; p < in + sb.st_size;)
    {   p = next(p, in + sb.st_size, &sparced);
        if  (NULL == sparced.gencl) break;
        if  (flag_v)
            fprintf( stderr, "%s genclass=%s, inclass=%s, tail=%s\n"
                   , sparced.mode ? "On " : "Off", sparced.gencl, sparced.incl, sparced.tail
                   );
        if  (flag_v > 1)
        {   fprintf(stderr, "    Argc=%d, exclude=%d", argc, exclude(sparced.gencl));
            if  (argc > 1) fprintf(stderr, " genclass(%s==%s)=%s"
                                  , sparced.gencl, genclass
                                  , strcmp(sparced.gencl, genclass) ? "No " : "Yes"
                                  );
            if  (argc > 2) fprintf(stderr, " inclass(%s==%s)=%s"
                                  , sparced.incl, inclass
                                  , strcmp(sparced.incl, inclass) ? "No " : "Yes"
                                  );
            fprintf(stderr, "\n");
        }

        if  (  (argc == 1 && !exclude(sparced.gencl))
            || (argc == 2 && 0 == strcmp(sparced.gencl, genclass))
            || (  argc == 3
               && 0 == strcmp(sparced.gencl, genclass)
               && 0 == strcmp(sparced.incl, inclass)
            )  )
        {   unsigned dodoma;
            char *path;
            unsigned char docod[] = {0, 0, jrsto, jrsto, 0, jrsta, jrsto, jrest};

            dodoma = docod[doma | sparced.mode];
            if  (0 > asprintf(&path, "%s/%s/%s", startupdir, todo, sparced.gencl))
                                                                                err(EX_TEMPFAIL, NULL);
            if  (dodoma & jrest && (0 > access(path, X_OK))) dodoma = jrsto | jrsta;
            free(path);
            if  (flag_v > 1) fprintf(stderr, "Todo %03o\n", dodoma);
            if  (dodoma & jrsto) exup("stop", sparced.incl, sparced.tail);
            if  (dodoma & jrest) exup("restart", sparced.incl, sparced.tail);
            if  (dodoma & jrsta) exup("start", sparced.incl, sparced.tail);
    }   }
    return(EX_OK);
}
